/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#ifndef _ZCOLOR_H_
#define _ZCOLOR_H_

#include "ZSystem.h"

/// Helper class to deal with colors
class ZColor
{
	public:

				uint32	color;

						operator uint32()
						{
							return color;
						}

						ZColor(uint32 c = 0xFF000000):
							color(c)
						{
						}

						ZColor(const ZColor &from)
						{
							color = from.color;
						}

						ZColor(uint32 a, uint32 r, uint32 g, uint32 b) 
						{
							r = r*a; r = ((r+(r>>8)+1)>>8);
							g = g*a; g = ((g+(g>>8)+1)>>8);
							b = b*a; b = ((b+(b>>8)+1)>>8);
							color = (a<<24)|(r<<16)|(g<<8)|b;
						}

	inline		void	SetRGB8(uint32 a, uint32 r, uint32 g, uint32 b)
						{
							r = r*a; r = ((r+(r>>8)+1)>>8);
							g = g*a; g = ((g+(g>>8)+1)>>8);
							b = b*a; b = ((b+(b>>8)+1)>>8);
							color = (a<<24)|(r<<16)|(g<<8)|b;
						}

	inline		void	SetSRGB(float32 a, float32 r, float32 g, float32 b)
						{
							a *= 255.; r *= a; g *= a; b *= a;
							color = (((sint32)a)<<24)|
									(((sint32)r)<<16)|
									(((sint32)g)<< 8)|
									( (sint32)b     );
						}

	inline		void	GetRGB8(uint08 &a, uint08 &r, uint08 &g, uint08 &b)
						{
							a = (uint08) ((color>>24)&0xFF);
							r = (uint08)(((color>>16)&0xFF)*0xFF)/a;
							g = (uint08)(((color>> 8)&0xFF)*0xFF)/a;
							b = (uint08)(((color    )&0xFF)*0xFF)/a;
						}

	inline		void	GetSRGB(float32 &a, float32 &r, float32 &g, float32 &b)
						{
							a = ((float32)((color>>24)&0xFF));
							r = ((float32)((color>>16)&0xFF))/a;
							g = ((float32)((color>> 8)&0xFF))/a;
							b = ((float32)((color    )&0xFF))/a;
							a /= 0xFF;
						}


// helper functions to deal with complete buffers and raw data
// Usually you will need that to preprocess bitmaps.

							/// Convert to premultiplied alpha value space
static	inline	uint32	AlphaMultiply(uint32 color)
						{
							uint32 r = (color>>24)*((color>>16)&0xFF);
							uint32 g = (color>>24)*((color>> 8)&0xFF);
							uint32 b = (color>>24)*((color    )&0xFF);
							return	(color&0xFF000000)|
									(((r+(r>>8)+1)>>8)<<16)|
									(((g+(g>>8)+1)>>8)<< 8)|
									(((b+(b>>8)+1)>>8)    );
						}

							/// Convert from premultiplied alpha value space
static	inline	uint32	AlphaDivide(uint32 color)
						{
							uint32 r = (0xFF*((color>>16)&0xFF))/(color>>24);
							uint32 g = (0xFF*((color>> 8)&0xFF))/(color>>24);
							uint32 b = (0xFF*((color    )&0xFF))/(color>>24);
							return	(color&0xFF000000)|(r<<16)|(g<< 8)|(b);
						}

							/// Convert to premultiplied alpha value space
static	inline	void	AlphaMultiply(uint32 *buffer, sint32 count)
						{
							for(count--;count>=0;count--)
							{
								uint32 c = buffer[count];
								uint32 r = (c>>24)*((c>>16)&0xFF);
								uint32 g = (c>>24)*((c>> 8)&0xFF);
								uint32 b = (c>>24)*((c    )&0xFF);
								buffer[count] = 
										(c&0xFF000000)|
										(((r+(r>>8)+1)>>8)<<16)|
										(((g+(g>>8)+1)>>8)<< 8)|
										(((b+(b>>8)+1)>>8)    );
							}
						}

							/// Convert from premultiplied alpha value space
static	inline	void	AlphaDivide(uint32 *buffer, sint32 count)
						{
							for(count--;count>=0;count--)
							{
								uint32 c = buffer[count];
								uint32 r = (0xFF*((c>>16)&0xFF))/(c>>24);
								uint32 g = (0xFF*((c>> 8)&0xFF))/(c>>24);
								uint32 b = (0xFF*((c    )&0xFF))/(c>>24);
								buffer[count] = (c&0xFF000000)|(r<<16)|(g<< 8)|(b);
							}
						}

							/// Interpolate between two colors
static	inline	uint32	Interpolate(uint32 a, uint32 b, uint08 pos)
						{
							// interpolate correctly between two possibly transparent colors 
							// assumes premultiplied color value   
							uint32 ca=((a>>24)     )*(uint32)(0xFF-pos)+((b>>24)     )*(uint32)pos;
							uint32 cr=((a>>16)&0xFF)*(uint32)(0xFF-pos)+((b>>16)&0xFF)*(uint32)pos;
							uint32 cg=((a>> 8)&0xFF)*(uint32)(0xFF-pos)+((b>> 8)&0xFF)*(uint32)pos;
							uint32 cb=((a    )&0xFF)*(uint32)(0xFF-pos)+((b    )&0xFF)*(uint32)pos;
							return	(((ca+(ca>>8)+1)>>8)<<24)|
									(((cr+(cr>>8)+1)>>8)<<16)|
									(((cg+(cg>>8)+1)>>8)<< 8)|
									(((cb+(cb>>8)+1)>>8)    );
						}
};


#endif