/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZMatrix.h"
#include <math.h>

/// Constructor sets matrix to identity.
ZMatrix::ZMatrix():
	_rotate(0.),
	_scale(1.,1.),
	_translate(0.,0.),
	_skewing(0.,0.),
	_isidentity(false)
{
	MakeIdentity();
}

/** Set rotation.
	The rotation angle is to be given in radiants.
*/
void ZMatrix::SetRotation(float32 angle)
{
	_rotate = angle;
	Update();
}

/** Set scaling.
	Scaling is specified as two factors 
	which are applied horizontally and vertically..
	To double the size of a path you might call 
	the method like that: mymatrix.SetScaling(ZPoint(2,2));
*/
void ZMatrix::SetScaling(ZPoint factor)
{
	_scale = factor;
	Update();
}

/** Set translation.
	Translates the path by the specified
	vector.	Translations are specified in absolute 
	distances. To move the path to the right by 10 pixels
	you can call the method using: 
	mymatrix.SetTranslation(ZPoint(10,0));
*/
void ZMatrix::SetTranslation(ZPoint vector)
{
	_translate = vector;
	Update();
}

/** Set skewing.
	Skewing is specified using an angle. The 
	angles are in radiants.
*/
void ZMatrix::SetSkewing(ZPoint angles)
{
	_skewing = angles;
	Update();
}

///
ZPoint ZMatrix::operator*(const ZPoint &from) const
{
	if(!_isidentity) 
	{
		return ZPoint(
			mtx[0*3+0]*from.x + mtx[0*3+1]*from.y + mtx[0*3+2],
			mtx[1*3+0]*from.x + mtx[1*3+1]*from.y + mtx[1*3+2]
		);
	}
	else
	{
		return from;
	}
}

///
ZMatrix ZMatrix::operator*(const ZMatrix &from) const
{
	ZMatrix res;
	if(!_isidentity || !from._isidentity) 
	{
		for(sint32 i=0; i<3; i++)
		{
			for(sint32 j=0; j<3; j++)
			{
				res.mtx[i*3+j] = from.mtx[i*3+0]*mtx[0*3+j]+
						 		 from.mtx[i*3+1]*mtx[1*3+j]+
						 		 from.mtx[i*3+2]*mtx[2*3+j];
			}
		}
		res._isidentity = false;
	}
	return res;
}

///
ZMatrix &ZMatrix::operator*=(const ZMatrix &from)
{
	if(!_isidentity || !from._isidentity) 
	{
		ZMatrix old(*this);
		for(sint32 i=0; i<3; i++)
		{
			for(sint32 j=0; j<3; j++)
			{
				mtx[i*3+j] = from.mtx[i*3+0]*old.mtx[0*3+j]+
							 from.mtx[i*3+1]*old.mtx[1*3+j]+
							 from.mtx[i*3+2]*old.mtx[2*3+j];
			}
		}
		_isidentity = false;
	}
	return *this;
}

///
ZMatrix &ZMatrix::operator=(const ZMatrix &from)
{
	mtx[0] = from.mtx[0];
	mtx[1] = from.mtx[1];
	mtx[2] = from.mtx[2];
	mtx[3] = from.mtx[3];
	mtx[4] = from.mtx[4];
	mtx[5] = from.mtx[5];
	mtx[6] = from.mtx[6];
	mtx[7] = from.mtx[7];
	mtx[8] = from.mtx[8];

	_isidentity = from._isidentity;
	_skewing	= from._skewing;
	_scale		= from._scale;
	_rotate		= from._rotate;
	_translate	= from._translate;

	return *this;
}

/// Clear the matrix (set to identity) and all properties
void ZMatrix::Clear()
{
	_rotate = 0.;
	_scale.Set(1.,1.);
	_translate.Set(0.,0.);
	_skewing.Set(0.,0.);
	MakeIdentity();
}

void ZMatrix::MakeIdentity()
{

	mtx[0*3+0] = 1;
	mtx[0*3+1] = 0;
	mtx[0*3+2] = 0;

	mtx[1*3+0] = 0;
	mtx[1*3+1] = 1;
	mtx[1*3+2] = 0;

	mtx[2*3+0] = 0;
	mtx[2*3+1] = 0;
	mtx[2*3+2] = 1;

	_isidentity = true;
}

/** Computes determinant of this matrix.
	This is usually used to compute the inversion
	of a matrix, but might be useful for other purposes.
*/
float32 ZMatrix::Determinant()
{
	return mtx[3*0+0] * (mtx[3*1+1]*mtx[3*2+2] - mtx[3*2+1]*mtx[3*1+2])
		 - mtx[3*0+1] * (mtx[3*1+0]*mtx[3*2+2] - mtx[3*2+0]*mtx[3*1+2])
		 + mtx[3*0+2] * (mtx[3*1+0]*mtx[3*2+1] - mtx[3*2+0]*mtx[3*1+1]);
}

void ZMatrix::SwapRows(ZMatrix &A, sint32 row1, sint32 row2) 
{ 
    float32   tmp; 
    for(sint32 i=0; i<3; i++) { 
        tmp = A.mtx[row1*3 + i]; 
        A.mtx[row1*3 + i] = A.mtx[row2*3 + i]; 
        A.mtx[row2*3 + i] = tmp; 
    } 
} 

#if 0

/** Computes the inversion of this matrix.
	The result will replace the data in this matrix.
	If the determinant is zero, e.g. the inversion can
	no be computed, the matrix is set to its identity.
*/
void ZMatrix::Inverse()
{ 
    sint32	maxrow; 
    float32	maxval; 
    float32	val; 

    ZMatrix  I;
    maxval = mtx[0]; 
    maxrow = 0; 
    for(sint32 i=0; i<3; i++) 
	{ 
        // 
        // Find row with largest value at the diagnal. 
        // 
        maxval = mtx[i*3 + i]; 
        maxrow = i; 
        {for(sint32 j=i+1; j<3; j++) 
		{ 
            val = mtx[j*3 + i]; 
            if(fabs(val) > fabs(maxval)) 
			{ 
                maxval = val; 
                maxrow = j; 
            } 
        }} 
  

        // 
        // Swap the row with largest value with current row. 
        // 
        if(maxrow != i) 
		{ 
            SwapRows(*this, i, maxrow); 
            SwapRows(I, i, maxrow); 
        } 

        // 
        // Divide the entire current row with maxval to get a 1 on the 
        // diagonal. 
        // 
        {for(sint32 k=0; k<3; k++) 
		{ 
            mtx[i*3 + k] /= maxval; 
            I.mtx[i*3 + k] /= maxval; 
        }} 

        // 
        // Subtract current row from all other rows so their values before the 
        // diagnal go zero. 
        // 
        {for(sint32 j=i+1; j<3; j++) 
		{ 
            val = mtx[j*3 + i]; 
            for(sint32 k=0; k<3; k++) 
			{ 
                mtx[j*3 + k] -= mtx[i*3 + k] * val; 
                I.mtx[j*3 + k] -= I.mtx[i*3 + k] * val; 
            } 
        }} 
    } 
  

    // Finally substract values so that the original matrix becomes identity 
    {for(sint32 i=2; i>=0; i--) 
	{ 
        for(sint32 j=i-1; j>=0; j--) 
		{ 
            val = mtx[j*3 + i]; 
            for(sint32 k=0; k<3; k++) 
			{ 
                mtx[j*3 + k] -= mtx[i*3 + k] * val; 
                I.mtx[j*3 + k] -= I.mtx[i*3 + k] * val; 
            } 
        } 
    }} 

	*this = I;
} 

#else

/** Computes the inversion of this matrix.
	The result will replace the data in this matrix.
	If the determinant is zero, e.g. the inversion can
	no be computed, the matrix is set to its identity.
*/
void ZMatrix::Inverse()
{
	if(!_isidentity)
	{
		float32 det = Determinant();
		if(det==0) 
		{
			MakeIdentity();
		}
		else
		{
			ZMatrix m;
			
			m.mtx[3*0+0] =  ( mtx[3*1+1]*mtx[3*2+2] - mtx[3*1+2]*mtx[3*2+1] ) / det;
			m.mtx[3*0+1] = -( mtx[3*0+1]*mtx[3*2+2] - mtx[3*2+1]*mtx[3*0+2] ) / det;
			m.mtx[3*0+2] =  ( mtx[3*0+1]*mtx[3*1+2] - mtx[3*1+1]*mtx[3*0+2] ) / det;
                                                                       
			m.mtx[3*1+0] = -( mtx[3*1+0]*mtx[3*2+2] - mtx[3*1+2]*mtx[3*2+0] ) / det;
			m.mtx[3*1+1] =  ( mtx[3*0+0]*mtx[3*2+2] - mtx[3*2+0]*mtx[3*0+2] ) / det;
			m.mtx[3*1+2] = -( mtx[3*0+0]*mtx[3*1+2] - mtx[3*1+0]*mtx[3*0+2] ) / det;
                                                                                     
			m.mtx[3*2+0] =  ( mtx[3*1+0]*mtx[3*2+1] - mtx[3*2+0]*mtx[3*1+1] ) / det;
			m.mtx[3*2+1] = -( mtx[3*0+0]*mtx[3*2+1] - mtx[3*2+0]*mtx[3*0+1] ) / det;
			m.mtx[3*2+2] =  ( mtx[3*0+0]*mtx[3*1+1] - mtx[3*0+1]*mtx[3*0+3] ) / det;

			m._isidentity = false;
			
			operator=(m);
		}
	}
}
#endif

void ZMatrix::Update()
{
	if(!_isidentity)
	{
		MakeIdentity();
	}
	
	ZMatrix m;

	// SHEAR
	if(_skewing.x!=0 || _skewing.y!=0)
	{
		m.mtx[0*3+0] = 1;
		m.mtx[0*3+1] = (float32)tan(_skewing.x);
		m.mtx[0*3+2] = 0;

		m.mtx[1*3+0] = (float32)tan(_skewing.y);
		m.mtx[1*3+1] = 1;
		m.mtx[1*3+2] = 0;

		m.mtx[2*3+0] = 0;
		m.mtx[2*3+1] = 0;
		m.mtx[2*3+2] = 1;

		_isidentity = false;

		operator*=(m);
	}

	// SCALE
	if(_scale.x!=1 || _scale.y!=1)
	{
		m.mtx[0*3+0] = _scale.x;
		m.mtx[0*3+1] = 0;
		m.mtx[0*3+2] = 0;

		m.mtx[1*3+0] = 0;
		m.mtx[1*3+1] = _scale.y;
		m.mtx[1*3+2] = 0;

		m.mtx[2*3+0] = 0;
		m.mtx[2*3+1] = 0;
		m.mtx[2*3+2] = 1;

		_isidentity = false;

		operator*=(m);

	}

	// ROTATION
	if(_rotate!=0)
	{
		m.mtx[0*3+0] = (float32)+cos(_rotate);
		m.mtx[0*3+1] = (float32)-sin(_rotate);
		m.mtx[0*3+2] = 0;

		m.mtx[1*3+0] = (float32)+sin(_rotate);
		m.mtx[1*3+1] = (float32)+cos(_rotate);
		m.mtx[1*3+2] = 0;

		m.mtx[2*3+0] = 0;
		m.mtx[2*3+1] = 0;
		m.mtx[2*3+2] = 1;

		_isidentity = false;

		operator*=(m);
	}

	// TRANSLATE	
	if(_translate.x!=0 || _translate.y!=0)
	{
		m.mtx[0*3+0] = 1;
		m.mtx[0*3+1] = 0;
		m.mtx[0*3+2] = _translate.x;

		m.mtx[1*3+0] = 0;
		m.mtx[1*3+1] = 1;
		m.mtx[1*3+2] = _translate.y;

		m.mtx[2*3+0] = 0;
		m.mtx[2*3+1] = 0;
		m.mtx[2*3+2] = 1;

		_isidentity = false;

		operator*=(m);
	}
}
