/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZOutliner.h"
#include "ZLine.h"
#include <assert.h>

///
ZOutliner::ZOutliner():
	joint(JOINT_MITER),
	firstmoveto(false),
	newsegment (false),
	rad(2)
{
}

///
ZOutliner::~ZOutliner()
{
}

///
ZPath ZOutliner::ComputeOutline(const ZPath &from)
{
	firstmoveto = true;		// by default we assume a moveto state
	newsegment  = true;

	outpath.MakeEmpty();
	innshap.MakeEmpty();
	outshap.MakeEmpty();

	from.Iterate(this);		// Lets go!!
	return outpath;
}

void ZOutliner::NewSegment(ZPoint a, ZPoint b)
{
	// Ingore zero length parts of the segment
	if(a != b)
	{
		// Get current part of the segment
		ZLine curr(a,b);
		// Get head of current part of the segment with width rad
		ZLine head(curr.Head(rad));
		// Get tail of current part of the segment with width rad
		ZLine tail(curr.Tail(rad));
		
		if(joint == JOINT_NONE)
		{
			if(newsegment)
			{
				newsegment = false;
				outpath += outshap;
				outpath += innshap;

				innshap.MakeEmpty();
				outshap.MakeEmpty();

				// Everything in innshap is reversed
				innshap.InsertClose(0);
				outshap.AddMoveTo(head.P1);

				innshap.InsertLineTo(0,head.P0);
				outshap.AddLineTo(head.P1);
			}

			innshap.InsertLineTo(0,tail.P0);
			outshap.AddLineTo(tail.P1);

			oldhead = head;
			oldtail = tail;

			// FIXME
			// This code will look very different
			// from the standard joints
		}
		else
		{
			// Is this the start of a new segment?
			// In this case me might have to really
			// handle joint stuff
			if(newsegment)
			{
				// Was a new shape section of the path started
				if(firstmoveto)
				{
					// In this case we do not have a joint
					// to put yet. Just add the segment
					savemovetop0 = head.P0;
					savemovetop1 = head.P1;
					// Everything in innshap is reversed
					innshap.InsertClose(0);
					outshap.AddMoveTo(head.P1);

					firstmoveto = false;

					// Store starting values in case a
					// Close() will need to add additional
					// joint segments
					strhead = head;
					strtail = tail;
				}
				else
				{
					switch(joint)
					{
						case	JOINT_ROUND:
						case	JOINT_MITER:
								{
									// Get the orientation of the old tail 
									// against the new head
									float32 orientation = head.Orientation(oldtail.P0);
									if(orientation != 0)
									{
										// If it is less than zero we know that the P1 area is
										// the inner part otherwise its the P0 area
										if(orientation <= 0) 
										{
											// Add miter joint by computing the intersection
											// of the two lines
											innshap.InsertLineTo(0,ZLine(oldhead.P0,oldtail.P0).Intersection(ZLine(head.P0,tail.P0)));
										}
										else
										{ 
											// Add miter joint by computing the intersection
											// of the two lines
											outshap.AddLineTo(ZLine(oldhead.P1,oldtail.P1).Intersection(ZLine(head.P1,tail.P1)));
										}
									}
								}
								break;
						default:
						case	JOINT_BEVEL:
								{
									// Automatic, nothing to do.
								}
								break;
					}
				}

				// We went through it once,
				// should be okay afterwards
				newsegment = false;

				innshap.InsertLineTo(0,head.P0);
				outshap.AddLineTo(head.P1);
			}

			// In any case need to build the path
			innshap.InsertLineTo(0,tail.P0);
			outshap.AddLineTo(tail.P1);

			// We need this info for the
			// computation of the joints.
			oldhead = head;
			oldtail = tail;
		}
	}
}

///
void ZOutliner::Start()
{
	ZSketcher::Start();
}

///
void ZOutliner::End()
{
	ZSketcher::End();
	
	if(!ShapeClosed())
	{
		// Just in case add the correct commands
		// to finish the shapes
		
		outpath += outshap;
		outpath += innshap;

		// FIXME!
		// Add caps
	}

	innshap.MakeEmpty();
	outshap.MakeEmpty();
}

///
void ZOutliner::MoveTo(ZPoint point)
{
	if(!ShapeClosed())
	{
		// Just in case add the correct commands
		// to finish the shapes

		outpath += outshap;
		outpath += innshap;

		// FIXME!
		// Add caps
	}

	innshap.MakeEmpty();
	outshap.MakeEmpty();

	firstmoveto = true;

	ZSketcher::MoveTo(point);
}

///
void ZOutliner::Close()
{
	newsegment = true;
	ZSketcher::Close();

	// This stuff here just
	// makes sure that we
	// connect the last segment
	// with the first one.
	if(joint == JOINT_NONE)
	{
		outpath += outshap;
		outpath += innshap;
		innshap.MakeEmpty();
		outshap.MakeEmpty();
	}
	else
	{
		switch(joint)
		{
			case	JOINT_ROUND:
			case	JOINT_MITER:
					{			
						float32 orientation = strhead.Orientation(oldtail.P0);
						if(orientation != 0)
						{
							// If it is less than zero we know that the P1 area is
							// the inner part otherwise its the P0 area
							if(orientation <= 0) 
							{
								// Add miter joint by computing the intersection
								// of the two lines
								innshap.InsertLineTo(0,ZLine(oldhead.P0,oldtail.P0).Intersection(ZLine(strhead.P0,strtail.P0)));
							}
							else 	
							{
								// Add miter joint by computing the intersection
								// of the two lines
								outshap.AddLineTo(ZLine(oldhead.P1,oldtail.P1).Intersection(ZLine(strhead.P1,strtail.P1)));
							}
						}
					}
					break;
			default:
			case	JOINT_BEVEL:
					{
						// Automatic, nothing to do.
					}
					break;
		}
		// Finally really close the
		// two paths which form the
		// stroke
		innshap.InsertMoveTo(0,savemovetop0);
		outshap.AddClose();

		outpath += outshap;
		outpath += innshap;

		innshap.MakeEmpty();
		outshap.MakeEmpty();
	}
}

///
void ZOutliner::LineTo(ZPoint point)
{
	newsegment = true;
	ZSketcher::LineTo(point);
}

///
void ZOutliner::CurveTo(ZPoint control, ZPoint point)
{
	newsegment = true;
	ZSketcher::CurveTo(control,point);
}

///
void ZOutliner::CurveTo(ZPoint control1, ZPoint control2, ZPoint point)
{
	newsegment = true;
	ZSketcher::CurveTo(control1,control2,point);
}

///
void ZOutliner::ArcTo(ZPoint radius, ZPoint point, float32 rotation, sint32 sweep_arc)
{
	newsegment = true;
	ZSketcher::ArcTo(radius,point,rotation,sweep_arc);
}
