/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZPath.h"

#include <string.h>

// defines the linear buffer increase value
#define ZBLOCKSIZE 	256

///
enum
{
	Z_CMD_NOP		= 0,
	Z_CMD_MOVETO 	= 1,
	Z_CMD_LINETO 	= 2,
	Z_CMD_QCURVETO 	= 3,
	Z_CMD_CCURVETO 	= 4,
	Z_CMD_ARCTO		= 5,
	Z_CMD_CLOSE		= 6,
	Z_STREAMID		= ('Z'<<24) | ('P'<<16) | ('T'<<8) | 'H'
};

///
ZPath::ZPath():
	_data(0),
	_used_size(0),
	_allc_size(0)
{
}

///
ZPath::ZPath(const ZPath &from):
	_data(0),
	_used_size(0),
	_allc_size(0)
{
	operator=(from);
}

///
ZPath::~ZPath()
{
	delete [] _data;
}

///
sint32 ZPath::StreamByteSize()
{
	return 2*sizeof(uint32)+_used_size*sizeof(uint32);
}

///
void *ZPath::ToStream(void *stream)
{
	uint32 *s = (uint32 *)stream;
	*s++ = Z_STREAMID;
	*s++ = _used_size;
	for(sint32 c=0;c<_used_size;c++)
		*s++ = _data[c];
	return s;
}

///
void *ZPath::FromStream(void *stream)
{
	uint32 *s = (uint32 *)stream;
	if(*s++ == Z_STREAMID)
	{
		delete [] _data;
		_data = 0;
		_allc_size = _used_size = 0;

		uint32 used_size = *s++;
		if(used_size)
		{
			_allc_size = _used_size = used_size;
			_data = new uint32 [_used_size];
			for(sint32 c=0;c<_used_size;c++)
				_data[c] = *s++;
		}
		return s;
	}
	return stream;
}

///
void ZPath::MakeEmpty()
{
	delete [] _data;
	_data = 0;
	_used_size = 0;
	_allc_size = 0;
}

///
void ZPath::AddMoveTo(ZPoint point)
{
	uint32 *dst = Add(3);
	*dst++ = Z_CMD_MOVETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::AddLineTo(ZPoint point)
{
	uint32 *dst = Add(3);
	*dst++ = Z_CMD_LINETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::AddCurveTo(ZPoint control, ZPoint point)
{
	uint32 *dst = Add(5);
	*dst++ = Z_CMD_QCURVETO;
	*((float32 *)dst++) = (float32)control.x;
	*((float32 *)dst++) = (float32)control.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::AddCurveTo(ZPoint control1, ZPoint control2, ZPoint point)
{
	uint32 *dst = Add(7);
	*dst++ = Z_CMD_CCURVETO;
	*((float32 *)dst++) = (float32)control1.x;
	*((float32 *)dst++) = (float32)control1.y;
	*((float32 *)dst++) = (float32)control2.x;
	*((float32 *)dst++) = (float32)control2.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::AddArcTo(ZPoint radius, ZPoint point, float32 rotation, sint32 large_arc, sint32 sweep_arc)
{
	uint32 *dst = Add(8);
	*dst++ = Z_CMD_ARCTO;
	*((float32 *)dst++) = (float32)radius.x;
	*((float32 *)dst++) = (float32)radius.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	*((float32 *)dst++) = (float32)rotation;
	*((sint32 *)dst++) = (sint32)large_arc;
	*((sint32 *)dst++) = (sint32)sweep_arc;
}

///
void ZPath::AddClose()
{
	uint32 *dst = Add(1);
	*dst++ = Z_CMD_CLOSE;
}

///
void ZPath::RemoveItem(sint32 index)
{
	uint32 *head = _data;
	uint32 *tail = _data;
	for(sint32 c=0;c<index;c++)
	{
		head = Next(head);
		// Vanilla save check
		if(head==&_data[_used_size])
		{
			// Woha, just return, the index was wrong...
			return;
		}
	}
	tail = Next(head);
	memmove(head,tail,(_used_size-(tail-_data))*sizeof(uint32));
	_used_size -= tail-head;
}

///
void ZPath::InsertMoveTo(sint32 index, ZPoint point)
{
	uint32 *dst = Insert(index,3);
	*dst++ = Z_CMD_MOVETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::InsertLineTo(sint32 index, ZPoint point)
{
	uint32 *dst = Insert(index,3);
	*dst++ = Z_CMD_LINETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::InsertCurveTo(sint32 index, ZPoint control, ZPoint point)
{
	uint32 *dst = Insert(index,5);
	*dst++ = Z_CMD_QCURVETO;
	*((float32 *)dst++) = (float32)control.x;
	*((float32 *)dst++) = (float32)control.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::InsertCurveTo(sint32 index, ZPoint control1, ZPoint control2, ZPoint point)
{
	uint32 *dst = Insert(index,7);
	*dst++ = Z_CMD_CCURVETO;
	*((float32 *)dst++) = (float32)control1.x;
	*((float32 *)dst++) = (float32)control1.y;
	*((float32 *)dst++) = (float32)control2.x;
	*((float32 *)dst++) = (float32)control2.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
}

///
void ZPath::InsertArcTo(sint32 index, ZPoint radius, ZPoint point, float32 rotation, sint32 large_arc, sint32 sweep_arc)
{
	uint32 *dst = Insert(index,8);
	*dst++ = Z_CMD_ARCTO;
	*((float32 *)dst++) = (float32)radius.x;
	*((float32 *)dst++) = (float32)radius.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	*((float32 *)dst++) = (float32)rotation;
	*((sint32 *)dst++) = (sint32)large_arc;
	*((sint32 *)dst++) = (sint32)sweep_arc;
}

///
void ZPath::InsertClose(sint32 index)
{
	uint32 *dst = Insert(index,1);
	*dst++ = Z_CMD_CLOSE;
}

///
void ZPath::ReplaceByMoveTo(sint32 index, ZPoint point)
{
	uint32 *dst = Replace(index,3);
	if(dst)
	{
		*dst++ = Z_CMD_MOVETO;
		*((float32 *)dst++) = (float32)point.x;
		*((float32 *)dst++) = (float32)point.y;
	}
}

///
void ZPath::ReplaceByLineTo(sint32 index, ZPoint point)
{
	uint32 *dst = Replace(index,3);
	if(dst)
	{
		*dst++ = Z_CMD_LINETO;
		*((float32 *)dst++) = (float32)point.x;
		*((float32 *)dst++) = (float32)point.y;
	}
}

///
void ZPath::ReplaceByCurveTo(sint32 index, ZPoint control, ZPoint point)
{
	uint32 *dst = Replace(index,5);
	if(dst)
	{
		*dst++ = Z_CMD_QCURVETO;
		*((float32 *)dst++) = (float32)control.x;
		*((float32 *)dst++) = (float32)control.y;
		*((float32 *)dst++) = (float32)point.x;
		*((float32 *)dst++) = (float32)point.y;
	}
}

///
void ZPath::ReplaceByCurveTo(sint32 index, ZPoint control1, ZPoint control2, ZPoint point)
{
	uint32 *dst = Replace(index,7);
	if(dst)
	{
		*dst++ = Z_CMD_CCURVETO;
		*((float32 *)dst++) = (float32)control1.x;
		*((float32 *)dst++) = (float32)control1.y;
		*((float32 *)dst++) = (float32)control2.x;
		*((float32 *)dst++) = (float32)control2.y;
		*((float32 *)dst++) = (float32)point.x;
		*((float32 *)dst++) = (float32)point.y;
	}
}

///
void ZPath::ReplaceByArcTo(sint32 index, ZPoint radius, ZPoint point, float32 rotation, sint32 large_arc, sint32 sweep_arc)
{
	uint32 *dst = Replace(index,8);
	if(dst)
	{
		*dst++ = Z_CMD_ARCTO;
		*((float32 *)dst++) = (float32)radius.x;
		*((float32 *)dst++) = (float32)radius.y;
		*((float32 *)dst++) = (float32)point.x;
		*((float32 *)dst++) = (float32)point.y;
		*((float32 *)dst++) = (float32)rotation;
		*((sint32 *)dst++) = (sint32)large_arc;
		*((sint32 *)dst++) = (sint32)sweep_arc;
	}
}

///
void ZPath::ReplaceByClose(sint32 index)
{
	uint32 *dst = Replace(index,1);
	if(dst)
	{
		*dst++ = Z_CMD_CLOSE;
	}
}

uint32 *ZPath::Add(sint32 count)
{
	ExtendIfNeeded(count);
	uint32 *dst = &_data[_used_size];
	_used_size += count;
	return dst;
}

uint32 *ZPath::Insert(sint32 index, sint32 count)
{
	uint32 *dst;
	
	dst = _data;
	{for(sint32 c=0;c<index;c++) 
	{
		dst=Next(dst); if(dst==&_data[_used_size]) break;
	}}

	ExtendIfNeeded(count);

	dst = _data;
	{for(sint32 c=0;c<index;c++) 
	{
		dst=Next(dst); if(dst==&_data[_used_size]) break;
	}}

	sint32 mamount = (_used_size-(dst-_data));
	if(mamount > 0)
	{
		memmove(&dst[count],dst,mamount*sizeof(uint32));
	}

	_used_size += count;
	return dst;
}

uint32 *ZPath::Replace(sint32 index, sint32 count)
{
	uint32 *head;
	uint32 *tail;
	
	head = _data;
	{for(sint32 c=0;c<index;c++) 
	{
		head=Next(head); 
		if(head==&_data[_used_size])
		{
			return 0;
		}
	}}
	tail = Next(head);

	uint32 diff = count-(tail-head);
	ExtendIfNeeded(diff);

	head = _data;
	{for(sint32 c=0;c<index;c++) 
	{
		head=Next(head); 
		if(head==&_data[_used_size])
		{
			return 0;
		}
	}}
	tail = Next(head);

	sint32 mamount = (_used_size-(tail-_data));
	if(mamount > 0 && diff !=0 )
	{
		memmove(&head[count],&head[count-diff],mamount*sizeof(uint32));
	}

	_used_size += count;
	return head;
}

uint32 *ZPath::Next(uint32 *src)
{
	switch(*src++)
	{
		case	Z_CMD_CLOSE:
		case	Z_CMD_NOP:
				break;
		case	Z_CMD_MOVETO:
				src+=2;
				break;
		case	Z_CMD_LINETO:
				src+=2;
				break;
		case	Z_CMD_QCURVETO:
				src+=4;
				break;
		case	Z_CMD_CCURVETO:
				src+=6;
				break;
		case	Z_CMD_ARCTO:
				src+=7;
				break;
		default:
				break;
	}
	return src;
}

///
void ZPath::Iterate(Iterator *iterator) const
{
	if(iterator)
	{
		ZPoint p0,p1,p2;
		float32 rotation;
		sint32 sweep_arc;
		sint32 large_arc;
		iterator->Start();
		for(uint32 *iter=_data;iter<(_data+_used_size);)
		{
			switch(*iter++)
			{
				case	Z_CMD_NOP:
						break;
				case	Z_CMD_MOVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						iterator->MoveTo(p0);
						break;
				case	Z_CMD_LINETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						iterator->LineTo(p0);
						break;
				case	Z_CMD_QCURVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						iterator->CurveTo(p0,p1);
						break;
				case	Z_CMD_CCURVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						p2.x = *((float32 *)iter++);
						p2.y = *((float32 *)iter++);
						iterator->CurveTo(p0,p1,p2);
						break;
				case	Z_CMD_ARCTO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						rotation = *((float32 *)iter++);
						large_arc = *((sint32 *)iter++);
						sweep_arc = *((sint32 *)iter++);
						iterator->ArcTo(p0,p1,rotation,large_arc,sweep_arc);
				case	Z_CMD_CLOSE:
						iterator->Close();
						break;
				default:
						return;
			}
		}
		iterator->End();
	}
}

void ZPath::ExtendIfNeeded(sint32 addsize)
{
	bool realloc = false;
	for(;(_used_size+addsize)>_allc_size;)
	{
		_allc_size += ZBLOCKSIZE;
		realloc = true;
	}
	if(realloc)
	{
		uint32 *newdata = new uint32 [_allc_size];
		if(_data)
		{
			memcpy(newdata,_data,_used_size*sizeof(uint32));
			delete [] _data;
		}
		_data = newdata;
	}
}


ZPath &ZPath::operator+=(const ZPath &from)
{
	ExtendIfNeeded(from._used_size);
	uint32 *dst = &_data[_used_size];
	memcpy(dst,from._data,from._used_size*sizeof(uint32));
	_used_size += from._used_size;
	return *this;
}

ZPath &ZPath::operator=(const ZPath &from)
{
	delete [] _data;
	_data = 0;
	_allc_size = _used_size = 0;
	operator+=(from);
	return *this;
}
