/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZSketcher.h"
#include <math.h>
#include <stdio.h>

/** A new path is about to be sampled.
	Implement this virtual method when you need to initialize
	things before a new path is sampled.
*/
void ZSketcher::Start()
{
	_shape_closed = true;
}

/** Path was finished sampling.
*/
void ZSketcher::End()
{
}

/**	Add a sampled segment to the output.
	When inhereting implement this virtual method to obtain the 
	sampled curves. In fact sampling is done by subdividing the curves 
	into linear segments. The sampling methods simply call NewSegment() with
	their sampled values. By default this method is empty.
*/
void ZSketcher::NewSegment(ZPoint a, ZPoint b)
{
}

/** Set the sampling rate for curves.
	Modifing this value can have great imact on performance. 
	While primitives should use a high value between 24 and 32,
	it is useful to lower this value for f.ex. truetype glyphs.
	Experience shows that a value of 8 is well suited for that.
*/
void ZSketcher::SetSamplingSteps(sint32 steps) 
{
	_samp_steps = steps;
	_step_pow_1 = 1.f/(float32)_samp_steps;
	_step_pow_2 = _step_pow_1*_step_pow_1;
	_step_pow_3 = _step_pow_2*_step_pow_1;
}

void ZSketcher::LineTo(ZPoint point)
{
	NewSegment(_curr_point,point);
	_curr_point = point;
}

void ZSketcher::MoveTo(ZPoint point)
{
	if(_shape_closed)
	{
		_strt_point = point;
		_shape_closed = false;
	}
	_curr_point = point;
}


void ZSketcher::CurveTo(ZPoint control, ZPoint point)
{
	// Compute forward difference values for a quadratic
	// curve of type A*(1-t)^2 + 2*B*t*(1-t) + C*t^2

	ZPoint f(_curr_point);
	ZPoint tmp((_curr_point-control*2.+point)*(float32)_step_pow_2);
	ZPoint ddf(tmp*2.);
	ZPoint df(tmp+(control-_curr_point)*(float32)(2.*_step_pow_1));

	for(sint32 c=0;c<(_samp_steps-1);c++)
	{
		f += df; df += ddf; 
		NewSegment(_curr_point,f);
		_curr_point = f;
	}

	// We specify the last point manually since
	// we obtain rounding errors during the
	// forward difference computation.
	NewSegment(_curr_point,point);
	_curr_point = point;
}

void ZSketcher::CurveTo(ZPoint control1, ZPoint control2, ZPoint point)
{
	float32 pre1 = 3*_step_pow_1;
	float32 pre2 = 3*_step_pow_2;
	float32 pre3 = _step_pow_3;
	float32 pre4 = 6*_step_pow_2;
	float32 pre5 = 6*_step_pow_3;
	
	ZPoint tmp1(_curr_point-control1*2+control2);
	ZPoint tmp2((control1-control2)*3-_curr_point+point);
	
	ZPoint f(_curr_point);
	ZPoint df((control1-_curr_point)*pre1+tmp1*pre2+tmp2*pre3);
	ZPoint ddf(tmp1*pre4+tmp2*pre5);
	ZPoint dddf(tmp2*pre5);
	
	for(sint32 c=0;c<(_samp_steps-1);c++)
	{
		f += df; df += ddf; ddf += dddf;		
		NewSegment(_curr_point,f);
		_curr_point = f;
	}

	NewSegment(_curr_point,point);
	_curr_point = point;
}

void ZSketcher::ArcTo(ZPoint radius, ZPoint point, float32 rotation, sint32 large_arc, sint32 sweep_arc)
{
	ZPoint A(_curr_point);
	ZPoint B(point);

	// Compute center
	ZPoint C((A+B)/2);
	{
		ZPoint D((A-C)/radius);
		ZPoint Cprim(-D.y,D.x);
		float32 dlen = D.Length();
		Cprim /= dlen;
		Cprim *= sqrt(1-dlen*dlen);
		Cprim *= radius;
		C+=Cprim;
	}

	// Compute starting and ending angles

	float32 angle0 = atan2((A-C).y/radius.y,(A-C).x/radius.x);
	float32 angle1 = atan2((B-C).y/radius.y,(B-C).x/radius.x);
 
	if(angle0 > angle1)
	{
		angle1 += 2*ZPI;
	}

	for(float32 a=angle0+.01;a<angle1;a+=.01)
	{
		ZPoint p(
			C.x + radius.x * cos(a),
			C.y + radius.y * sin(a)
		);
		NewSegment(_curr_point,p);
		_curr_point = p;
	}

	NewSegment(_curr_point,point);
	_curr_point = point;

/*	ZPoint mid((point+_curr_point)/2);

	mid /= radius;

	float32 a = sqrt(1 - pow(mid.Length(),2));

	ZPoint cen((mid*a)/mid.Length()/2);

	cen *= radius;

	NewSegment(_curr_point,cen+_curr_point);
	NewSegment(cen+_curr_point,point);*/
}

void ZSketcher::Close()
{
	// Connect start point with end point
	NewSegment(_curr_point,_strt_point);
	_curr_point = _strt_point;
	_shape_closed = true;
}

/*
	Computation of the quadratic forward difference values

	Generic form:
	A(1-t)^2 + 2Bt(1-t) + C(t^2)

	Isolate t:
	= A(1+t^2-2t) + 2Bt - 2B(t^2) + C(t^2)
	= A + A(t^2) - 2At + 2Bt - 2B(t^2) + C(t^2)
	= (A - 2B + C)(t^2) + (-2A + 2B)t + A

	Substitute factors:
	a = (A - 2B + C)
	b = (-2A + 2B)
	c = A

	Generic quadratic curve:
	f(t) = a(t^2) + bt + c

    First derivation:
	f(t+dt)-f(t) = a((t+dt)^2) + b(t+dt) + c - (a(t^2) + bt + c)
	= a(t^2+dt^2+2tdt) + bt + bdt + c - a(t^2) - bt - c
	= a(t^2) + a(dt^2) + a2tdt + bdt - a(t^2)
	= a(dt^2) + a2tdt + bdt

	df(t) = a2tdt + a(dt^2) + bdt

    Second derivation:
	df(t+dt)-df(t) = a2dt(t+dt) + a(dt^2) + bdt - (a2tdt + a(dt^2) + bdt)
	= 2a(dt^2)

	ddf(t) = 2a(dt^2);

    Result:
	f(0) = A
	df(0) = (A - 2B + C)*dt^2 + 2*(-A + B)*dt
	ddf(0) = 2(A - 2B + C)*dt^2
*/
