/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZStencil.h"
#include "ZTools.h"
#include <math.h>

// Just a note:
// All default values in the constructors
// are meant to be used for the SPACE_PATH
// (normalized object) coordinate space.

ZStencil::ZStencil(void):
	space(SPACE_PATH)
{
}

ZStencil::~ZStencil()
{
}

ZLinearGradientStencil::ZLinearGradientStencil()
{
	SetVector(ZLine(0,0,0,1.));
}

/** Set vector of this linear gradient.
	The gradient will be mapped on this vector.
*/
void ZLinearGradientStencil::SetVector(const ZLine &v)
{
	vect = v;

	ZPoint s(v.P1-v.P0);
	float32 f = 1/s.Length();
	ZMatrix mtx0; 
	mtx0.SetTranslation(-v.P0);
	vecm  = mtx0;
	mtx0.Clear();
	mtx0.SetRotation((float32)-atan2(s.y,s.x));
	vecm *= mtx0;
	mtx0.Clear();
	mtx0.SetScaling(ZPoint(f,f));
	vecm *= mtx0;
}

void ZLinearGradientStencil::SetPosition(sint32 x, sint32 y)
{
	ZPoint p0((float32)x+0.f,(float32)y);
	ZPoint p1((float32)x+1.f,(float32)y);

	// Convert pixel to selected space
	// if necessary
	if(Space()==SPACE_PATH)
	{
		p0 = (((p0*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
		p1 = (((p1*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
	}

	// Apply custom stencil matrix
	// Apply vector
	p0 *= StencilMatrix()*vecm;
	p1 *= StencilMatrix()*vecm;

	sint32 in0=(sint32)(p0.x*65536*16);
	sint32 in1=(sint32)(p1.x*65536*16);
	
	curp = in0;
	incc = in1-in0;
}
                                                                            
uint32 ZLinearGradientStencil::GetPixel(bool skip)
{
#if ZRENDERER_TOPBOTTOM
	if(!skip)
#endif
	{
		uint32 color = GetColor(curp>>(20-ZGRADIENTBUFFERBITS)); 
		curp+=incc;
		return color;
	}
#if ZRENDERER_TOPBOTTOM
	curp+=incc;
	return 0;
#endif
}

ZRadialGradientStencil::ZRadialGradientStencil():
	center(.5f,.5f),
	radius(.5f),
	focus(.5f,.5f)
{
}

/** Set the properties for the radial gradient.
	Center and radius defines largest/outermost circle for the 
	radial gradient.
*/
void ZRadialGradientStencil::SetVector(const ZPoint &center, float32 radius)
{
	SetVector(center,radius,center);
}

/** Set the properties for the radial gradient.
	Center and radius defines largest/outermost circle for the 
	radial gradient. Focus defines the focal point for the radial gradient.
*/
void ZRadialGradientStencil::SetVector(const ZPoint &c, float32 r, const ZPoint &f)
{
	center = c;
	radius = r;
	focus  = f;
}

void ZRadialGradientStencil::SetPosition(sint32 x, sint32 y)
{
	ZPoint p0((float32)x+0.f,(float32)y);
	ZPoint p1((float32)x+1.f,(float32)y);
	ZPoint fc(focus);

	// Convert pixel to selected space
	if(Space()==SPACE_PATH)
	{
		p0 = (((p0*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
		p1 = (((p1*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
	}

	// Apply custom stencil matrix
	p0 *= StencilMatrix();
	p1 *= StencilMatrix();
	fc *= StencilMatrix();

	// Prepare for real usage
	// Its pointless to create a seperate 
	// matrix for that...
	p1 -= center;
	p0 -= center;
	fc -= center;
	p1 /= radius;
	p0 /= radius;
	fc /= radius;

#if 0

	float32 fx = fc.x;
	float32 fy = fc.y;
	float32 px = p0.x;
	float32 py = p0.y;
	float32 dx = px-fx;
	float32 dy = py-fy;

	float32 t = (p1.x-p0.x);

	b = dx*fx + dy*fy;
	db = t*fx;

	rad = b*b + dx*dx + dy*dy;
	drad = 2*t*p0.x + t*t + db*(db + 2*fx*px + 2*fy*py - 2*fy*fy - 2*fx*fx - 2);
	ddrad = 2*t*t + 2*db*db;

#else
	i = p1-p0;
	p = p0;
	f = fc;
#endif
}

uint32 ZRadialGradientStencil::GetPixel(bool skip)
{
#if ZRENDERER_TOPBOTTOM
//	if(!skip)
#endif
	{
#if 0
		uint32 color = GetColor((sint32)(((b+sqrt(rad))/2)*1024));
		b += db;
		rad += drad;
		drad += ddrad;
		return color;
#else
		float32 dx = (p.x-f.x);
		float32 dy = (p.y-f.y);
		float32 cr = 1/(1-f.x*f.x-f.y*f.y);
		b = (dx*f.x+dy*f.y)*cr;
		rad = b*b+(dx*dx + dy*dy)*cr;
		uint32 color = GetColor((sint32)(((b+sqrt(rad)))*1024));
		p+=i;
		return color;
#endif
	}
#if ZRENDERER_TOPBOTTOM
	return 0;
#endif
}

ZBitmapStencil::ZBitmapStencil():
	offset(0.,0.),
	size(1.,1.)
{ 
	for(sint32 i=0;i<ZMIMMAPLEVEL;i++) mipb[i]=0;
	bmp = 0;
}

ZBitmapStencil::~ZBitmapStencil() 
{ 
	for(sint32 i=0;i<ZMIMMAPLEVEL;i++) delete [] mipb[i];
}

void ZBitmapStencil::SetVector(const ZPoint &o, const ZPoint &s)
{
	offset = o;
	size = s;
}

/**	Set new bitmap pattern to be used for rendering.
	Bitmap should be in ARGB 32 bit mode with premultiplied alpha values. 
	Please note that this method will make a private copy of the pattern,
	which means that you can free your buffer at any time after this call.
	In fact this is done since this method anyway creates a mip map for the
	pattern.
*/
void ZBitmapStencil::SetBitmap(const uint32 *buffer, uint32 width, uint32 height)
{
	for(sint32 i=0;i<ZMIMMAPLEVEL;i++) 
	{	
		delete [] mipb[i];
		mipb[i]=0;
	}
	if(buffer && width && height)
	{

		mipb[0] = new uint32[width * height];
		mipx[0] = width;
		mipy[0] = height;
		memcpy(mipb[0],buffer,width * height * sizeof(uint32));
		for(sint32 c=1;c<ZMIMMAPLEVEL;c++)
		{
			mipx[c] = mipx[c-1]/2;
			mipy[c] = mipy[c-1]/2;
			if(mipx[c]==0 || mipy[c]==0)
			{
				// FIXME: This here is really ugly...
				sint32 sc=0;
				sint32 sa=0;
				sint32 sr=0;
				sint32 sg=0;
				sint32 sb=0;
				for(uint32 y=0;y<mipy[c-1];y++)
				{
					for(uint32 x=0;x<mipx[c-1];x++)
					{ 
						sc++;
						uint32 p = mipb[c-1][y*mipx[c-1]+x];
						sa += (p>>24)&0xFF;
						sr += (p>>16)&0xFF;
						sg += (p>> 8)&0xFF;
						sb += (p    )&0xFF;
					}
				}
				mipb[c] = new uint32[1];
				mipx[c] = 1;
				mipy[c] = 1;
				mipb[c][0] = 
					((sa/sc)<<24)|
					((sr/sc)<<16)|
					((sg/sc)<< 8)|
					((sb/sc)    );
			}
			else
			{
				mipb[c]  = new uint32[mipx[c]*mipy[c]];
				for(uint32 y=0;y<mipy[c];y++)
				{
					for(uint32 x=0;x<mipx[c];x++)
					{
						uint32 p0 = mipb[c-1][((y*2)+0)*mipx[c-1]+(x*2)+0];
						uint32 p1 = mipb[c-1][((y*2)+0)*mipx[c-1]+(x*2)+1];
						uint32 p2 = mipb[c-1][((y*2)+1)*mipx[c-1]+(x*2)+0];
						uint32 p3 = mipb[c-1][((y*2)+1)*mipx[c-1]+(x*2)+1];
						mipb[c][y*mipx[c]+x] = 
							(((((p0>>24)&0xFF) + ((p1>>24)&0xFF) + ((p2>>24)&0xFF) + ((p3>>24)&0xFF))/4)<<24)|
							(((((p0>>16)&0xFF) + ((p1>>16)&0xFF) + ((p2>>16)&0xFF) + ((p3>>16)&0xFF))/4)<<16)|
							(((((p0>> 8)&0xFF) + ((p1>> 8)&0xFF) + ((p2>> 8)&0xFF) + ((p3>> 8)&0xFF))/4)<< 8)|
							(((((p0    )&0xFF) + ((p1    )&0xFF) + ((p2    )&0xFF) + ((p3    )&0xFF))/4)    );
					}
				}
			}
		}
	}
}

void ZBitmapStencil::SetPosition(sint32 x, sint32 y)
{
	ZPoint p0((float32)x+0.f,(float32)y);
	ZPoint p1((float32)x+1.f,(float32)y);

	// Convert pixel to selected space
	// if necessary
	if(Space()==SPACE_PATH)
	{
		p0 = (((p0*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
		p1 = (((p1*PathMatrix())-PathFrame().Origin())/PathFrame().Size());
	}

	// Apply custom stencil matrix
	p0 *= StencilMatrix();
	p1 *= StencilMatrix();

	p0 -= offset;
	p1 -= offset;
	p0 /= size;
	p1 /= size;

	sint32 ix0=(sint32)(p0.x*65536*16);
	sint32 ix1=(sint32)(p1.x*65536*16);

	curx = ix0;
	incx = ix1-ix0;

	sint32 iy0=(sint32)(p0.y*65536*16);
	sint32 iy1=(sint32)(p1.y*65536*16);

	cury = iy0;
	incy = iy1-iy0;

	// compute mip map level
	if(bmp==0)
	{
		sint32 levl=0;
		for(float32 iter=(p1-p0).Length()/2;iter>=1;iter/=2.) levl++;
		levl= ZMIN(ZMIMMAPLEVEL-1,levl);

		// set correct mip map
		bmp = mipb[levl];
		wdh = mipx[levl];
		hgt = mipy[levl];
	}

	curx *= wdh;
	cury *= hgt;
	incx *= wdh;
	incy *= hgt;
}
                                                                            
uint32 ZBitmapStencil::GetPixel(bool skip)
{
#if ZRENDERER_TOPBOTTOM
	if(!skip)
#endif
	{
		if(bmp)
		{
			uint32 cx = (curx>>20)+(wdh<<12);
			uint32 cy = (cury>>20)+(hgt<<12);

			uint32 x0 = ((cx  )%wdh); // Uaaah!!!
			uint32 y0 = ((cy  )%hgt);
			uint32 x1 = ((x0+1)%wdh);
			uint32 y1 = ((y0+1)%hgt);

			// Get the 4 affected pixels;
			uint32 in_pix0 = bmp[y0*wdh+x0];
			uint32 in_pix1 = bmp[y0*wdh+x1];
			uint32 in_pix2 = bmp[y1*wdh+x0];
			uint32 in_pix3 = bmp[y1*wdh+x1];

			if((in_pix0>>24) || (in_pix1>>24) || (in_pix2>>24) || (in_pix3>>24))
			{
				// Compute bilinear interpolation
				sint32  mx1 = (curx>>8)&0xFFF;
				sint32  mx0 = 4096-mx1;
				sint32  my1 = (cury>>8)&0xFFF;
				sint32  my0 = 4096-my1;

				uint32 m1 = (my0*mx0);
				uint32 m2 = (my0*mx1);
				uint32 m3 = (my1*mx0);
				uint32 m4 = (my1*mx1);

				// Actually apply the factors
				uint32 a = (m1*((in_pix0>>24)&0xFF) + m2*((in_pix1>>24)&0xFF) + m3*((in_pix2>>24)&0xFF) + m4*((in_pix3>>24)&0xFF))>>24;
				uint32 r = (m1*((in_pix0>>16)&0xFF) + m2*((in_pix1>>16)&0xFF) + m3*((in_pix2>>16)&0xFF) + m4*((in_pix3>>16)&0xFF))>>24;
				uint32 g = (m1*((in_pix0>> 8)&0xFF) + m2*((in_pix1>> 8)&0xFF) + m3*((in_pix2>> 8)&0xFF) + m4*((in_pix3>> 8)&0xFF))>>24;
				uint32 b = (m1*((in_pix0    )&0xFF) + m2*((in_pix1    )&0xFF) + m3*((in_pix2    )&0xFF) + m4*((in_pix3    )&0xFF))>>24;

				curx+=incx;
				cury+=incy;
				return (a<<24) | (r<<16) | (g<< 8) | (b    );
			}
		}
	}
#if ZRENDERER_TOPBOTTOM
	curx+=incx;
	cury+=incy;
	return 0;
#endif
}
