/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#ifndef _ZSTENCIL_H_
#define _ZSTENCIL_H_

#include "ZMatrix.h"
#include "ZGradient.h"
#include "ZLine.h"
#include "ZRect.h"

/// Abtract stencil class used to obtain fill of paths
class ZStencil
{
	public:

		enum
		{
			SPACE_PATH,
			SPACE_VIEW
		};

						ZStencil(void);
		virtual			~ZStencil();

							/// Set stencil matrix
				void	SetStencilMatrix(const ZMatrix &matrix) { smtx = matrix; smtx.Inverse(); }
							/// Return stencil matrix;
		const	ZMatrix	&StencilMatrix() const { return smtx; }
							/// Set coordinate space the stecil will work in
				void	SetSpace(sint32 mode) { space = mode; }
							/// Obtain coordinate space
		const	sint32	Space() const { return space; }

	protected:	

		friend class ZRenderer;

		const	ZRect	&PathFrame() const { return pframe; }
		const 	ZMatrix &PathMatrix() const { return pmtx; }

		virtual	void	SetPosition(sint32 x, sint32 y) = 0;
		virtual	uint32	GetPixel(bool skip) = 0;
		virtual	bool	IsMonotone() const = 0;
		virtual	void	Reset() { };

	private:


		inline	void	SetPathFrame(const ZRect &rect) { pframe = rect; }
		inline	void	SetPathMatrix(const ZMatrix &matrix) { pmtx = matrix; pmtx.Inverse(); }

				ZRect	pframe;
				ZMatrix	pmtx;
				ZMatrix smtx;
				sint32	space;
};

/// Implements plain color stencil
class ZColorStencil : public ZStencil
{
	public:

						ZColorStencil(uint32 c = 0xFFFFFFFF):color(c) {}
		virtual			~ZColorStencil() {}

							/// Set color in ARGB 32 bit format
				void	SetColor(uint32 c) { color = c; }
							/// Obtain color of this color stencil in ARGB 32 bit format
		const	uint32	Color() const { return color; }

	protected:

		virtual	void	SetPosition(sint32 x, sint32 y) { };
		virtual	uint32	GetPixel(bool skip) { return color; }
		virtual	bool	IsMonotone() const { return true; }

	private:
				uint32	color;
};

/// Implements linear gradient fill stencil
class ZLinearGradientStencil : public ZStencil, public ZGradient
{
	public:

						 ZLinearGradientStencil();
		virtual			~ZLinearGradientStencil() { }

				void	SetVector(const ZLine &v);
		const	ZLine	&Vector() const { return vect; } 

	protected:

		virtual	void	SetPosition(sint32 x, sint32 y);
		virtual	uint32	GetPixel(bool skip);
		virtual bool	IsMonotone() const { return false; }

	private:
				
				ZLine	vect;
				ZMatrix vecm;
				sint32	curp;
				sint32	incc;
};

/// Implements radial gradient fill stencil
class ZRadialGradientStencil : public ZStencil, public ZGradient
{
	public:
						ZRadialGradientStencil();
		virtual			~ZRadialGradientStencil() { }

				void	SetVector(const ZPoint &center, float32 radius);
				void	SetVector(const ZPoint &center, float32 radius, const ZPoint &focus);

		const	ZPoint	&Center() const { return center; }
		const	float32	Radius() const { return radius; }
		const	ZPoint	&Focus() const { return focus; }
	
	protected:
			
		virtual	void	SetPosition(sint32 x, sint32 y);
		virtual	uint32	GetPixel(bool skip);
		virtual bool	IsMonotone() const { return false; }

	private:
				ZPoint	center;
				float32	radius;
				ZPoint	focus;

				float32	b;
				float32	db;
				float32	rad;
				float32	drad;
				float32	ddrad;

				ZPoint	p;
				ZPoint	f;
				ZPoint	i;
};

/// Implements bitmap fill stencil
class ZBitmapStencil : public ZStencil
{
	public:

						ZBitmapStencil();
		virtual			~ZBitmapStencil();

				void	SetBitmap(const uint32 *buffer, uint32 width, uint32 height);

				void	SetVector(const ZPoint &offset, const ZPoint &size);
			
	protected:

		virtual	void	SetPosition(sint32 x, sint32 y);
		virtual	uint32	GetPixel(bool skip);
		virtual bool	IsMonotone() const { return false; }
		virtual	void	Reset() { bmp = 0; };

	private:
				ZPoint	offset;
				ZPoint	size;

				uint32	*mipb[ZMIMMAPLEVEL];
				uint32	mipx[ZMIMMAPLEVEL];
				uint32	mipy[ZMIMMAPLEVEL];

				uint32	*bmp;
				uint32	wdh;
				uint32	hgt;
				
				sint32	curx;
				sint32	incx;
				sint32	cury;
				sint32	incy;
};

#endif
