/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which is included in this distribution for details.
*/
#include "ZXML.h"
#include <iostream>
#include <fstream>

// check if next character in stream is a whitespace
static bool next_whitespace(iostream &stream)
{
	switch(stream.peek())
	{
		case	0x0A:
		case	0x0D:
		case	0x20:
		case	'\t':
				return true;
		default:
				return false;
	}
}

// skip all whitespace character in current stream
static void skip_whitespace(iostream &stream)
{
	for(;;)
	{
		if(stream.eof() || !stream.good())
			return;

		if(next_whitespace(stream))
			stream.get();
		else
			return;
	}
}

bool ZXMLAttribute::Parse(iostream &stream, ZXMLIterator *iter)
{
	erase();
	name.erase();

	skip_whitespace(stream);

	// check for closing
	if((stream.peek() == '/') || (stream.peek() == '>')) 
		return false;

	// obtain attribute name
	for(;;)
	{
		if(stream.eof() || !stream.good())
			return false;

		if(!next_whitespace(stream) && stream.peek() != '=')
		{
			if(
				(stream.peek() == '/') ||
				(stream.peek() == '>') 
			) throw ZXMLParseError("XML Attribute not set.");
			name += stream.get();
		}
		else
		{
			break;
		}
	}

	skip_whitespace(stream);

	// obtain attribute content
	if(stream.get()=='=')
	{
		skip_whitespace(stream);

		if(stream.get()=='"')
		{
			char c;
			for(;(c=stream.get())!='"';)
			{
				if(stream.eof() || !stream.good())
					return false;

				*this+=c;
			}
			return true;
		}
	}
	throw ZXMLParseError("XML Attribute not set.");
	return false;
}

bool ZXMLElement::Parse(iostream &stream, ZXMLIterator *iter)
{
	name.erase();
	clear();
	att.clear();
//	body.clear();

	skip_whitespace(stream);

	if(stream.get()=='<')
	{
		// Check for closing
		if(stream.peek() == '/')
			throw ZXMLParseError("XML Element not opened");

		// Skip comments or extra data
		if(stream.peek() == '!' || stream.peek() == '?')
		{
			for(;stream.get()!='>';) 
			{
				if(stream.eof() || !stream.good())
					return false;
			}
			return false;
		}

		// Obtain element name
		for(;;)
		{
			if(stream.eof() || !stream.good())
				return false;

			if(!next_whitespace(stream))
			{
				if(
					(stream.peek() == '/') ||
					(stream.peek() == '>') 
				) break;

				name += stream.get();
			}
			else
			{
				break;
			}
		}

		iter->ElementNamed(*this);

		skip_whitespace(stream);

		// Parse attributes
		ZXMLAttribute a;
		for(;a.Parse(stream,iter);)
		{
			att.push_back(a);
		}

		iter->ElementAttributed(*this);

		skip_whitespace(stream);

		// Check for single element
		if(stream.peek()=='/')
		{
			stream.get();
			if(stream.peek()=='>')
			{
				stream.get();
				iter->ElementCompleted(*this);
				return true;
			}
		}

		// We have a body?
		if(stream.peek()=='>')
		{
			stream.get();
			ZXMLElement e;
			for(;;)
			{
				if(stream.eof() || !stream.good())
					return false;

				string curbody;
				for(;stream.peek()!='<';)
				{
					curbody += stream.get();
				}
//				body.push_back(curbody);
				stream.get();
				if(stream.peek()=='/')
				{
					stream.get();
					skip_whitespace(stream);
					string checkstring;
					for(;;)
					{
						if(stream.eof() || !stream.good())
							return false;

						if(!next_whitespace(stream) && (stream.peek() != '>'))
						{
							checkstring += stream.get();
						}
						else
						{
							break;
						}
					}
					if(checkstring!=name)
					{
						throw ZXMLParseError("XML Element not closed.");
					}
					if(next_whitespace(stream))
					{
						skip_whitespace(stream);
					}
					if(stream.peek()!='>')
					{
						throw ZXMLParseError("Parse error.");
					}
					stream.get();
					iter->ElementCompleted(*this);
					return true;
				}
				stream.seekp(-1,ios::cur);
				if(e.Parse(stream,iter))
				{
					push_back(e);
				}
			}
		}
		throw ZXMLParseError("Parse error.");
	}
	return false;
}

void ZXMLDocument::Parse(string filename, ZXMLIterator *iter)
{
	fstream fs;
	fs.open(filename.c_str(),ios::in|ios::binary);
	if(fs.good())
	{
		try
		{
			for(;fs.peek()!='<' && (!fs.eof());) 
			{ 
				fs.get(); 
			}
		}

		catch(ZXMLParseError error)
		{
//			OutputDebugString(error.Error().c_str());
//			OutputDebugString("\n");
		}

		catch(...)
		{
		}

		ZXMLElement e;
		for(;!fs.eof();)
		{
			if(e.Parse(fs,iter))
			{
				root.push_back(e);
			}
		}

		fs.close();
	}
}
